/*
 * Copyright Altera Corporation (C) 2013. All rights reserved
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <linux/module.h>
#include <linux/ide.h>
#include <linux/platform_device.h>
#include <linux/of_irq.h>
#include <linux/of_address.h>

#include "altera_rpde.h"
#include "altera_epde.h"

#define MAJOR_NR                240
#define DEVICE_NAME             "blkrpde"

struct block_device_t {
	spinlock_t     lock;
	void __iomem   *rpdata;
	unsigned long  rpdata_avoff;
	unsigned long  datlen;
	void __iomem   *dmacsr;
	void __iomem   *dmades;
	unsigned int   rphwirq;
	void __iomem   *epdata;
	unsigned long  epdata_avoff;
	char           *sys_buf;
	unsigned long  sys_buf_phys;
	void __iomem   *perfcounter;
	unsigned long  timediff;
	unsigned long  perfcountertick;
	unsigned int   dmaxfer_cmd;
	struct completion dma_done;
	struct gendisk *gd;
};
static struct block_device_t block_device;

/* Support only one DES */
static void init_msgdma_xfer(void __iomem *dmades, unsigned long frphy,
			     unsigned long tophy, u32 size32)
{
	writel(frphy, dmades + MSGDMA_READ_ADDR);
	writel(tophy, dmades + MSGDMA_WRITE_ADDR);
	writel(size32, dmades + MSGDMA_DES_LEN);

	/* record start time */
	/* reset */
	writel(PERFCTR_GBL_CLK_CTR_LO_RST,
		block_device.perfcounter + PERFCTR_GBL_CLK_CTR_LO);
	/* start */
	writel(PERFCTR_GBL_CLK_CTR_HI_START,
		block_device.perfcounter + PERFCTR_GBL_CLK_CTR_HI);

	/* start transfer */
	writel(MSGDMA_DES_CTL_TX_CH   |
	       MSGDMA_DES_CTL_TC_IRQ  |
	       MSGDMA_DES_CTL_GO,
	      dmades + MSGDMA_DES_CTL);
}

/* Simple SGDMA csr with interrupt */
static void init_msgdma_csr(void __iomem *dmacsr)
{
	/* Stop and clear CSR */
	MSGDMA_STOP_CTL(dmacsr);

	/* Clear any SDGMA STS */
	MSGDMA_CLR_STS(dmacsr);

	/* Enable CSR */
	MSGDMA_ENA_CTL(dmacsr);
}

/* MSGDMA transfer function
 *   return timediff from performance counter
 */
static unsigned long msgdma_transfer(unsigned long frphy, unsigned long tophy,
				     u32 size32)
{
	unsigned long rt;

	/* Setup performance counter and barrier */
	block_device.perfcountertick = 0;
	init_completion(&block_device.dma_done);

	/* Setting up SGDMA CSR */
	init_msgdma_csr(block_device.dmacsr);

	/* Init SGDMA transfer */
	init_msgdma_xfer(block_device.dmades, frphy, tophy, size32);

	/* wait DMA transfer complete */
	wait_for_completion(&block_device.dma_done);

	rt = (block_device.perfcountertick / RP_PERFCOUNTER_MFREQ);  /* usec */

	return rt;
}

static int block_ioctl(struct block_device *bdev, fmode_t mode,
		       unsigned int cmd, unsigned long arg)
{
	int rt;
	struct ioctl_t ioctl_data;
	unsigned long sys_buf_avoff;

	block_device.rpdata_avoff = RP_DMA2OCR_OFF;
	sys_buf_avoff = block_device.sys_buf_phys + RP_DMA2SDRAM_OFF;

	switch (cmd) {
	case GET_IOCTL:
		return copy_to_user((void *)arg,
				    (void *)&block_device.datlen,
				    sizeof(long));
	case SET_IOCTL:
		return copy_from_user((void *)&block_device.dmaxfer_cmd,
					  (void __user *)arg,
				      sizeof(int));
	case PCI_RX_IOCTL:
		/* From EP On-Chip RAM to RP On-Chip RAM using SGDMA */
		memset(block_device.rpdata, 0, block_device.datlen);
		ioctl_data.datlen = block_device.datlen;
		ioctl_data.timediff = msgdma_transfer(block_device.epdata_avoff,
						      block_device.rpdata_avoff,
						      block_device.datlen);
		rt = copy_to_user((void *)arg, (void *)&ioctl_data,
				  sizeof(struct ioctl_t));
		return rt;
	case PCI_TX_IOCTL:
		/* From RP On-Chip RAM to EP On-Chip RAM using SGDMA */
		memset(block_device.epdata, 0, block_device.datlen);
		ioctl_data.datlen = block_device.datlen;
		ioctl_data.timediff = msgdma_transfer(block_device.rpdata_avoff,
						      block_device.epdata_avoff,
						      block_device.datlen);
		rt = copy_to_user((void *)arg, (void *)&ioctl_data,
				  sizeof(struct ioctl_t));
		return rt;
	case SYS_RX_IOCTL:
		/* From EP On-Chip RAM to RP System Memory using SGDMA */
		memset(block_device.sys_buf, 0, block_device.datlen);
		ioctl_data.datlen = block_device.datlen;
		ioctl_data.timediff = msgdma_transfer(block_device.epdata_avoff,
						      sys_buf_avoff,
						      block_device.datlen);
		rt = copy_to_user((void *)arg, (void *)&ioctl_data,
				  sizeof(struct ioctl_t));
		return rt;
	case SYS_TX_IOCTL:
		/* From RP System Memory to EP On-Chip RAM using SGDMA */
		memset(block_device.epdata, 0, block_device.datlen);
		ioctl_data.datlen = block_device.datlen;
		ioctl_data.timediff = msgdma_transfer(sys_buf_avoff,
						      block_device.epdata_avoff,
						      block_device.datlen);
		rt = copy_to_user((void *)arg, (void *)&ioctl_data,
				  sizeof(struct ioctl_t));
		return rt;
	}

	return -ENOTTY;
}

static struct request_queue *block_queue;
static const struct block_device_operations block_fops = {
	.owner = THIS_MODULE,
	.ioctl = block_ioctl,
};

static void blk_request(struct request_queue *rq)
{
	int err;
	struct request *req;
	unsigned long offset, nbytes;

	req = blk_fetch_request(rq);
	while (req) {
		if (req->cmd_type != REQ_TYPE_FS)
			continue;

		offset = blk_rq_pos(req) * SECTOR_SIZE;
		nbytes = blk_rq_cur_sectors(req) * SECTOR_SIZE;
		err = -EIO;
		if ((offset + nbytes) <= block_device.datlen) {
			err = 0;
			switch (rq_data_dir(req)) {
			case READ:
				switch (block_device.dmaxfer_cmd) {
				case PCI_RX_IOCTL:
					/* Read from RP On-Chip RAM to Buffer */
					memcpy(req->buffer,
						block_device.rpdata + offset,
						nbytes);
					break;
				case SYS_RX_IOCTL:
					/* Read from System Buffer to Buffer */
					memcpy(req->buffer,
						block_device.sys_buf + offset,
						nbytes);
					break;
				case PCI_TX_IOCTL:
				case SYS_TX_IOCTL:
					/* Read from EP On-Chip RAM to Buffer */
					memcpy(req->buffer,
						block_device.epdata + offset,
						nbytes);
					break;
				}
				break;
			case WRITE:
				switch (block_device.dmaxfer_cmd) {
				case PCI_RX_IOCTL:
				case SYS_RX_IOCTL:
					/* Write from Buffer to EP OCM */
					memcpy(block_device.epdata + offset,
						req->buffer, nbytes);
					break;
				case PCI_TX_IOCTL:
					/* Write from Buffer to RP OCM */
					memcpy(block_device.rpdata + offset,
						req->buffer, nbytes);
					break;
				case SYS_TX_IOCTL:
					/* Write from Buffer to System Buffer */
					memcpy(block_device.sys_buf + offset,
						req->buffer, nbytes);
					break;
				}
				break;
			default:
				err = -EIO;
				pr_notice("Unknown request %u\n",
				       rq_data_dir(req));
				break;
			}
		}

		if (__blk_end_request_cur(req, err) == 0)
			req = blk_fetch_request(rq);
	}
}

static int block_device_init(struct platform_device *pdev)
{
	int err;

	/* register as block device */
	err = register_blkdev(MAJOR_NR, DEVICE_NAME);
	if (err) {
		dev_err(&pdev->dev, "block device fail to register\n");
		goto err_ret;
	}
	spin_lock_init(&block_device.lock);

	block_queue = blk_init_queue(blk_request, &block_device.lock);
	if (!block_queue) {
		dev_err(&pdev->dev, "block_device: error in blk_init_queue\n");
		goto err_unregdev;
	}

	/* whole disk, no partition */
	block_device.gd = alloc_disk(1);
	if (!block_device.gd) {
		dev_err(&pdev->dev, "block_device: error in alloc_disk\n");
		goto err_deinit_queue;
	}

	block_device.gd->major = MAJOR_NR;
	block_device.gd->first_minor = 0;
	block_device.gd->fops = &block_fops;
	block_device.gd->private_data = &block_device;
	sprintf(block_device.gd->disk_name, "%s%d", DEVICE_NAME, 0);
	set_capacity(block_device.gd,
		    (block_device.datlen/SECTOR_SIZE));

	block_device.gd->queue = block_queue;

	add_disk(block_device.gd);
	return 0;
err_deinit_queue:
	blk_cleanup_queue(block_queue);
err_unregdev:
	unregister_blkdev(MAJOR_NR, DEVICE_NAME);
err_ret:
	return err;
}

static irqreturn_t rpde_isr(int irq, void *arg)
{
	const u32 *ptr1, *ptr2;

	switch (block_device.dmaxfer_cmd) {
	case PCI_RX_IOCTL:
	case PCI_TX_IOCTL:
		ptr1 = (u32 *)block_device.rpdata;
		ptr2 = (u32 *)block_device.epdata;
		break;
	case SYS_RX_IOCTL:
	case SYS_TX_IOCTL:
		ptr1 = (u32 *)block_device.sys_buf;
		ptr2 = (u32 *)block_device.epdata;
		break;
	default:
		/* Unknow command */
		return IRQ_NONE;
	}

	/* last 4 bytes location */
	ptr1 += ((block_device.datlen / 4) - 1);
	ptr2 += ((block_device.datlen / 4) - 1);
	/* wait until last 4 bytes posted write complete */
	while (*ptr1 != *ptr2);

	/* Stop counter */
	__raw_writel(PERFCTR_GBL_CLK_CTR_LO_STOP,
		block_device.perfcounter + PERFCTR_GBL_CLK_CTR_LO);

	/* Clear Interrupt and RUN bits */
	__raw_writel((__raw_readl(block_device.dmacsr + MSGDMA_CTL)
				  | MSGDMA_CTL_DIS_BITS)
				  & ~MSGDMA_CTL_IE_GLOBAL,
		     block_device.dmacsr + MSGDMA_CTL);

	if (__raw_readl(block_device.dmacsr + MSGDMA_STS) & MSGDMA_STS_IRQ)
		/* read tick */
		block_device.perfcountertick = readl(block_device.perfcounter +
						     PERFCTR_GBL_CLK_CTR_LO);

	/* Clear any SDGMA STS */
	__raw_writel(MSGDMA_STS_CLR_BITS, block_device.dmacsr + MSGDMA_STS);

	/* completed */
	complete(&block_device.dma_done);

	return IRQ_HANDLED;
}

/*
 * Map (physical) PCI mem into (virtual) kernel space
 */
static void __iomem *remap_pci_mem(ulong base, ulong size)
{
	ulong page_base = ((ulong) base) & PAGE_MASK;
	ulong page_offs = ((ulong) base) - page_base;
	void __iomem *page_remapped = ioremap_nocache(page_base,
						      page_offs + size);

	return page_remapped ? (page_remapped + page_offs) : NULL;
}

static const struct of_device_id altr_sgdma_ids[] = {
	{ .compatible = "ALTR,msgdma-1.0", },
	{ .compatible = "ALTR,msgdma-dispatcher-1.0", },
	{},
};

static int altr_rpde_pci_probe(struct platform_device *pdev)
{
	int err = -EIO;
	struct pci_dev *lpdev = NULL;
	struct device_node *node = NULL;
	struct resource res;
	block_device.timediff = 0;

	/* Max block length */
	block_device.datlen = min3((unsigned long)RP_OCRAM_SIZE,
				   (unsigned long)EP_OCRAM_SIZE,
				   (unsigned long)SGDMA_MAX_SIZE);

	/*********** Is the right EP available? ***********/
	lpdev = pci_get_device(PCI_VENDOR_ID_EPDE, PCI_DEVICE_ID_EPDE, NULL);
	if (!lpdev) {
		dev_err(&pdev->dev, "has no require EP detected\n");
		err = -ENODEV;
		return err;
	}
	if ((lpdev->subsystem_vendor != PCI_SUBVEN_ID_EPDE) ||
	    (lpdev->subsystem_device != PCI_SUBDEV_ID_EPDE)) {
		dev_err(&pdev->dev, "has no require EP detected\n");
		err = -ENODEV;
		goto err_nodev;
	}

	/*********** RP OCRAM Setup ***********/
	res.start = RP_OCRAM_SBASE;
	res.end = res.start + RP_OCRAM_SIZE - 1;
	res.flags = IORESOURCE_MEM;
	res.name = "RP OCR";

	block_device.rpdata = devm_request_and_ioremap(&pdev->dev, &res);
	if (!block_device.rpdata) {
		dev_err(&pdev->dev, "failed to mapping onchip memory region\n");
		goto err_nodev;
	}

	/* Performance counter */
	node = of_find_compatible_node(NULL, NULL, "altr,perf-counter-1.0");
	if (of_address_to_resource(node, 0, &res)) {
		dev_err(&pdev->dev, "Missing performance counter register\n");
		goto err_nodev;
	}
	block_device.perfcounter = devm_request_and_ioremap(&pdev->dev, &res);
	if (!block_device.perfcounter) {
		dev_err(&pdev->dev,
			"failed to mapping performance counter region\n");
		goto err_nodev;
	}

	/*********** SGDMA Setup ***********/
	node = of_find_matching_node(NULL, altr_sgdma_ids);
	if (!node) {
		dev_err(&pdev->dev, "No SGDMA found in Device Tree\n");
		goto err_nodev;
	}

	/* SGDMA CSR */
	if (of_address_to_resource(node, 0, &res)) {
		dev_err(&pdev->dev, "Missing SGDMA CSR register\n");
		goto err_nodev;
	}
	block_device.dmacsr = devm_request_and_ioremap(&pdev->dev, &res);
	if (!block_device.dmacsr) {
		dev_err(&pdev->dev, "fail to mapping SGDMA CSR region\n");
		goto err_nodev;
	}

	/* SGDMA descriptor */
	if (of_address_to_resource(node, 1, &res)) {
		dev_err(&pdev->dev, "Missing SGDMA descriptor register\n");
		goto err_nodev;
	}
	block_device.dmades = devm_request_and_ioremap(&pdev->dev, &res);
	if (!block_device.dmades) {
		dev_err(&pdev->dev,
			"fail to mapping SGDMA descriptor region\n");
		goto err_nodev;
	}

	/* verify if SGDMA status ok */
	writel(~0UL, block_device.dmacsr + MSGDMA_STS);
	if ((readl(block_device.dmacsr + MSGDMA_STS) & MSGDMA_STS_MASK) !=
	    MSGDMA_STS_AT_RESET) {
		dev_err(&pdev->dev, "SGDMA fail status test\n");
		goto err_nodev;
	}

	/* SGDMA IRQ */
	block_device.rphwirq = irq_of_parse_and_map(node, 0);
	err = devm_request_irq(&pdev->dev, block_device.rphwirq, rpde_isr,
			0, pdev->name, &block_device);
	if (err) {
		dev_err(&pdev->dev, "failed to register IRQ\n");
		goto err_nodev;
	}

	/*********** Block Setup ***********/
	/* Init block device API */
	err = block_device_init(pdev);
	if (err) {
		dev_err(&pdev->dev, "fail to init block device\n");
		goto err_nodev;
	}

	/* Allocate system memory for DMA trasfer */
	block_device.sys_buf = dma_alloc_coherent(&pdev->dev,
		block_device.datlen, (dma_addr_t *)&block_device.sys_buf_phys,
		GFP_KERNEL | GFP_DMA);
	if (!block_device.sys_buf) {
		dev_err(&pdev->dev,
			"failed to allocate system memory for DMA\n");
		goto err_nodev;
	}

	block_device.epdata_avoff = RP_DMA2TXS_OFF + EP_OCRAM_BAR_AVOFF;
	block_device.epdata =
		remap_pci_mem(pci_resource_start(lpdev, EP_DIR_BAR_NR)
			      + EP_OCRAM_BAR_AVOFF,
			      block_device.datlen);
	if (!block_device.epdata) {
		dev_err(&pdev->dev, "failed to mapping epdata\n");
		goto err_nodev;
	}

	pci_dev_put(lpdev);
	return 0;
err_nodev:
	pci_dev_put(lpdev);
	return err;
}

static int altr_rpde_pci_remove(struct platform_device *pdev)
{
	dma_free_coherent(&pdev->dev, block_device.datlen,
		block_device.sys_buf, block_device.sys_buf_phys);
	iounmap(block_device.epdata);
	unregister_blkdev(MAJOR_NR, DEVICE_NAME);
	del_gendisk(block_device.gd);
	put_disk(block_device.gd);
	blk_cleanup_queue(block_queue);
	return 0;
}

static const struct of_device_id rpde_of_match[] = {
	{ .compatible = "altr,pcie-rpde", NULL },
	{ },
};
MODULE_DEVICE_TABLE(of, rpde_of_match);

static struct platform_driver rpde_driver = {
	.driver = {
		.name = "altera-rpde",
		.owner = THIS_MODULE,
		.of_match_table = of_match_ptr(rpde_of_match),
	},
	.probe = altr_rpde_pci_probe,
	.remove = altr_rpde_pci_remove,
};

static int __init altr_rpde_pci_init(void)
{
	return platform_driver_register(&rpde_driver);
}

static void __exit altr_rpde_pci_exit(void)
{
	platform_driver_unregister(&rpde_driver);
}

module_init(altr_rpde_pci_init);
module_exit(altr_rpde_pci_exit);

MODULE_DESCRIPTION("Altera RootPort Design Example with SGDMA Block Transfer");
MODULE_AUTHOR("Ley Foon Tan <lftan@altera.com>");
MODULE_LICENSE("GPL v2");
